/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducermp3click.h"
#include "../gdamsongdatabase.h"
#include <math.h>
#include <gtk/gtksignal.h>
#include <gdam/gdamsample.h>

static void     gdam_producer_mp3_click_real_set_filename
					   (GdamProducerMp3Click* mp3_click,
					    const char*      filename);
static void     gdam_producer_mp3_click_play     (GdamProducer*    producer,
					    GdamPlayInfo*    play_info);
static void     gdam_producer_mp3_click_destroy  (GtkObject*       object);

static GtkObjectClass* parent_class = NULL;

static void build_time_vols(gint64* time_vols,
			    gdouble beat_period,
			    gdouble click_volume);

enum {
	KEYPOINTS,
	LAST_ARG
};

static void beat_info_to_data(GdamBeatInfo* beat_info, double* beat_data,
			      gdouble position)
{
	gdouble bpm, first_beat;
	bpm = gdam_beat_info_get_bpm(beat_info, position);
	first_beat = gdam_beat_info_get_first_beat(beat_info, position);
	g_return_if_fail(bpm > 0.001);
	beat_data[0] = first_beat;
	beat_data[1] = 60.0 / bpm;
}

static void refresh_fade_array(GdamProducerMp3Click* mp3)
{
	GdamModel* model = GDAM_MODEL(mp3);
	gint64* time_vols;
	GdamArg arg[2];

	if (model->source_id) {
		GdamBeatInfo* beat_info;
		gdouble beat_data[2];
		beat_info = &model->beat_info;
		time_vols = alloca(sizeof(gint64) * 8);
		beat_info_to_data(beat_info, beat_data, 0);
		build_time_vols(time_vols, beat_data[1], mp3->click_volume);
		arg[0].name = "keypoints";
		arg[0].type = GDAM_ARG_TYPE_INT64_ARRAY;
		GDAM_ARG_INT64_ARRAY(arg+0).length = 8;
		GDAM_ARG_INT64_ARRAY(arg+0).array = time_vols;
		arg[1].name = "phase_sync";
		arg[1].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
		GDAM_ARG_DOUBLE_ARRAY(arg+1).length = 2;
		GDAM_ARG_DOUBLE_ARRAY(arg+1).array = beat_data;
		gdam_model_set_args_full(model, mp3->fader_id, KEYPOINTS, 
					 2, arg);
	}
}


static void gdam_producer_mp3_click_arg_set_volume(GtkObject* object, 
						   GdamArg* arg)
{
	GDAM_PRODUCER_MP3_CLICK(object)->click_volume = GDAM_ARG_DOUBLE(arg);
	refresh_fade_array(GDAM_PRODUCER_MP3_CLICK(object));
}

static void gdam_producer_mp3_click_arg_get_volume(GtkObject* object, 
						   GdamArg* arg)
{
	 GDAM_ARG_DOUBLE(arg) = GDAM_PRODUCER_MP3_CLICK(object)->click_volume;
}

static void gdam_producer_mp3_click_arg_set_filename(GtkObject* object, GdamArg* arg)
{
	gdam_producer_mp3_click_real_set_filename(GDAM_PRODUCER_MP3_CLICK(object),
	                                    GDAM_ARG_STRING(arg));
}
static void gdam_producer_mp3_click_arg_get_filename(GtkObject* object, GdamArg* arg)
{
	GdamProducerMp3Click* producer_mp3_click = GDAM_PRODUCER_MP3_CLICK(object);
	GDAM_ARG_STRING(arg) = g_strdup(producer_mp3_click->filename);
}

static gboolean gdam_producer_mp3_click_real_set_beatinfo (GdamModel* model,
							const GdamBeatInfo* new)
{
	gboolean rv = GDAM_MODEL_CLASS(parent_class)->set_beat_info(model, new);
	refresh_fade_array(GDAM_PRODUCER_MP3_CLICK(model));
	return rv;
}

static void gdam_producer_mp3_click_class_init(GdamProducerMp3ClickClass* klass)
{
	GdamProducerClass* producer_class = GDAM_PRODUCER_CLASS(klass);
	GdamModelClass* model_class = GDAM_MODEL_CLASS(klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	producer_class->play = gdam_producer_mp3_click_play;
	model_class->set_beat_info = gdam_producer_mp3_click_real_set_beatinfo;
	GTK_OBJECT_CLASS(klass)->destroy = gdam_producer_mp3_click_destroy;
	parent_class = gtk_type_class(GDAM_TYPE_PRODUCER);
	gdam_arg_register(object_class,
	                  "filename",
			  GDAM_ARG_TYPE_STRING,
			  gdam_producer_mp3_click_arg_set_filename,
			  gdam_producer_mp3_click_arg_get_filename,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "click_volume",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_producer_mp3_click_arg_set_volume,
			  gdam_producer_mp3_click_arg_get_volume,
			  GDAM_ARG_VITAL);
	model_class->num_arg_sets = LAST_ARG;
}

static void gdam_producer_mp3_click_init(GdamProducerMp3Click* producer_mp3_click)
{
	producer_mp3_click->fade_envelope = 64;
	producer_mp3_click->click_volume = 1.0;
}

GtkType gdam_producer_mp3_click_get_type()
{
	static GtkType mp3_click_type = 0;
	if (!mp3_click_type) {
		static const GtkTypeInfo mp3_click_info = {
			"GdamProducerMp3Click",
			sizeof(GdamProducerMp3Click),
			sizeof(GdamProducerMp3ClickClass),
			(GtkClassInitFunc) gdam_producer_mp3_click_class_init,
			(GtkObjectInitFunc) gdam_producer_mp3_click_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		mp3_click_type = gtk_type_unique(GDAM_TYPE_PRODUCER, &mp3_click_info);
	}
	return mp3_click_type;
}

GdamProducerMp3Click* gdam_producer_mp3_click_new(GdamChannel*	channel)
{
	GdamProducerMp3Click* rv;
	rv = GDAM_PRODUCER_MP3_CLICK(gtk_type_new(GDAM_TYPE_PRODUCER_MP3_CLICK));
	((GdamModel*)rv)->channel = channel;
	return rv;
}


static void mp3_click_ended_callback(gpointer user_data)
{
	GdamProducerMp3Click* mp3_click = GDAM_PRODUCER_MP3_CLICK(user_data);
	gdam_producer_source_ended(GDAM_PRODUCER(mp3_click), 0.0);
}

static void eof_connect_success(int eof_sub_id, gpointer user_data)
{
	GdamProducerMp3Click* mp3_click = GDAM_PRODUCER_MP3_CLICK(user_data);
	mp3_click->eof_sub_id = eof_sub_id;
	gdam_channel_register_handler(
		GDAM_MODEL(mp3_click)->channel,
		gdam_channel_simple_handler_new(eof_sub_id,
						TRUE,
						mp3_click_ended_callback,
						mp3_click)
	);
}

static void build_time_vols(gint64* time_vols,
			    gdouble beat_period,
			    gdouble click_volume)
{
	int pulse_left = 50;
	int pulse_right = 300;
	int beat_length = (int)(44100.0 * beat_period);
	GdamFixed peak_vol;
	peak_vol = (GdamFixed)((double)GdamFixed_ONE * click_volume);
	time_vols[0] = 0;
	time_vols[1] = peak_vol;
	time_vols[2] = pulse_right;
	time_vols[3] = 0;
	time_vols[4] = beat_length - pulse_left;
	time_vols[5] = 0;
	time_vols[6] = beat_length;
	time_vols[7] = peak_vol;
}


static void got_mp3_subsource(int id, gpointer user_data)
{
	GDAM_PRODUCER_MP3_CLICK(user_data)->mp3_id = id;
}

static void didnt_get_mp3_subsource(gpointer user_data)
{
	GDAM_PRODUCER_MP3_CLICK(user_data)->mp3_id = 0;
	g_message("error creating mp3");
}

static void fader_success(int id, gpointer user_data)
{
	GDAM_PRODUCER_MP3_CLICK(user_data)->fader_id = id;
}

static void     gdam_producer_mp3_click_play       (GdamProducer*    producer,
					      GdamPlayInfo*    play_info)
{
	GdamProducerMp3Click* mp3_click = GDAM_PRODUCER_MP3_CLICK(producer);
	GdamChannel* channel;
	GdamModel* model;
	int num_args;
	int mp3_id, fader_id = 0, mix_id, wavetable_id;
	GdamArg args[4];
	GdamRequestBuilder* request_builder;
        gdouble  position = play_info ? play_info->position : 0.0;

	g_return_if_fail(mp3_click != NULL);
	if (!mp3_click->filename) {
		g_warning("gdam_producer_mp3_click_play: no filename");
		return;
	}
	model = GDAM_MODEL(mp3_click);
	if (model->create_source_pending) {
		g_warning("gdam_producer_mp3_click_play: already requesting");
		return;
	}
	channel = model->channel;
	if (!channel){
		g_warning("mp3_click model missing channel!");
		return;
	}

	mp3_click->mp3_id = 0;
	mp3_click->fader_id = 0;

	gdam_debug("producermp3_click: not in sync.", position);
	args[0].type = GDAM_ARG_TYPE_STRING;
	args[0].name = "filename";
	GDAM_ARG_STRING(args + 0) = mp3_click->filename;

	args[1].type = GDAM_ARG_TYPE_INT64;
	args[1].name = "position";
	GDAM_ARG_INT64(args + 1) = (gint64)(position * 44100.0);

	num_args = 2;
	if (mp3_click->buffer_size > 0) {
		args[num_args].type = GDAM_ARG_TYPE_INT;
		args[num_args].name = "buffer_size";
		GDAM_ARG_INT(args + num_args) = mp3_click->buffer_size;
		num_args++;
	}
	if (mp3_click->fade_envelope > 0) {
		args[num_args].type = GDAM_ARG_TYPE_INT;
		args[num_args].name = "fade_envelope";
		GDAM_ARG_INT(args + num_args) = mp3_click->fade_envelope;
		num_args++;
	}
	if (play_info->length > 0) {
		args[num_args].type = GDAM_ARG_TYPE_INT64;
		args[num_args].name = "length";
		GDAM_ARG_INT64(args + num_args) = 
				(gint64)(play_info->length * 44100.0);
		num_args++;
	}

	request_builder = gdam_request_builder_new();
	mp3_id = gdam_request_builder_append_create(
			request_builder,
			gdam_request_create_source_v(
				"GdamSeekableMp3",
				num_args,
				args,
				got_mp3_subsource,
				didnt_get_mp3_subsource,
				mp3_click));
	if (gdam_beat_info_get_bpm(&model->beat_info, position) > 0) {
		gint64* time_vols;
		/* XXX: make beat_info have num beat information */
		int num_beats = 100;
		int num_time_vols = (3 * num_beats + 2) * 2;
		gdouble beat_data[2];
		GdamBeatInfo* beat_info = &model->beat_info;

		wavetable_id = gdam_request_builder_append_create(
				request_builder,
				gdam_request_create_source(
					"GdamWavetable",
					NULL, NULL, NULL,
					4,
					"buffer_id",
					GDAM_ARG_TYPE_INT,
					1,
					"hertz",
					GDAM_ARG_TYPE_DOUBLE,
					440.0,
					"use_overall_position",
					GDAM_ARG_TYPE_INT,
					1,
					"position",
					GDAM_ARG_TYPE_INT64,
					(gint64)(position*44100.0)));
		
		time_vols = alloca(sizeof(gint64) * 8);
		beat_info_to_data(beat_info, beat_data, position);
		build_time_vols(time_vols,
		                beat_data[1],
				mp3_click->click_volume);
		g_message("num_time_vols: %d", num_time_vols);


		fader_id = gdam_request_builder_append_create(
				request_builder,
				gdam_request_create_source(
					"GdamFader",
					fader_success, 
					NULL, 
					mp3_click,
					4,
					"keypoints",
					GDAM_ARG_TYPE_INT64_ARRAY,
					8,
					time_vols,
					"mode",
					GDAM_ARG_TYPE_STRING,
					"loops",
					"filtered",
					GDAM_ARG_TYPE_INT,
					wavetable_id,
					"phase_sync",
					GDAM_ARG_TYPE_DOUBLE_ARRAY,
					2, beat_data));
	}
	{
		gint64 mixed[4];
		mixed[0] = 0;
		mixed[1] = mp3_id;
		mixed[2] = 0;
		mixed[3] = fader_id;

		mix_id = gdam_request_builder_append_create(
				request_builder,
				gdam_request_create_source(
					"GdamMix",
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					mp3_click,
					2,
					"mixed",
					GDAM_ARG_TYPE_INT64_ARRAY,
					fader_id == 0 ? 2 : 4,
					mixed,
					"mode",
					GDAM_ARG_TYPE_STRING,
					"first_ended"));
	}

	gdam_debug("Creating producing mp3_click with position %f", position);
	gtk_object_ref(GTK_OBJECT(mp3_click));
	model->create_source_pending = 1;
			
	gdam_request_builder_append_generic(
		request_builder,
		gdam_request_subscribe_source_ended(
			mix_id,
			eof_connect_success,
			NULL,
			mp3_click)
	);

	gdam_model_new_child_request(model->parent,
	                             model,
			             play_info,
				     mix_id,
				     request_builder);
}

static void gdam_producer_mp3_click_real_set_filename(GdamProducerMp3Click* mp3_click,
                                                const char*      filename)
{
	GdamSongDatabase* database = gdam_song_database_system();
	GdamSongInfo* info = NULL;
	GdamBeatInfo beat_info;
	gboolean was_playing = FALSE;
	gdouble position = 0;
	        /* TODO: might this permit multiple requests? */
	if (GDAM_MODEL(mp3_click)->source_id != 0){
		was_playing = TRUE;
		gdam_producer_stop(GDAM_PRODUCER(mp3_click));
	}
	if (mp3_click->filename)
		g_free(mp3_click->filename);
	mp3_click->filename = g_strdup(filename);
	if (database)
		info = gdam_song_database_find(database, filename);
	gdam_beat_info_set_first_beat(&beat_info, 0);
	gdam_beat_info_set_bpm(&beat_info, 90);    /* TODO: get rid of this */
	if (info) {
		gdam_producer_update_length(GDAM_PRODUCER(mp3_click), 
					    gdam_song_info_get_length(info));
			/* TODO: memory leak from get_bi? */
		position = gdam_beat_info_get_first_beat( 
					gdam_song_info_get_beat_info(info),0);
		beat_info = *(gdam_song_info_get_beat_info(info));
	}
	gdam_producer_set_position(GDAM_PRODUCER(mp3_click), position);
	gdam_model_set_beat_info(GDAM_MODEL(mp3_click), &beat_info);
}

static void gdam_producer_mp3_click_destroy(GtkObject* object)
{
	GdamProducerMp3Click* mp3_click = GDAM_PRODUCER_MP3_CLICK(object);
	g_free(mp3_click->filename);
	(*parent_class->destroy)(object);
}

void gdam_producer_mp3_click_set_filename(GdamProducerMp3Click*  	producer_mp3_click,
				    const char* 	filename)
{
	GdamArg arg;
	arg.name = "filename";
	arg.type = GDAM_ARG_TYPE_STRING;
	GDAM_ARG_STRING(&arg) = (char*) filename;
	gdam_arg_set_object_arg(GTK_OBJECT(producer_mp3_click), &arg);
}

void gdam_producer_mp3_click_set_buffer_size(GdamProducerMp3Click* producer_mp3_click,
                                       int              buffer_size)
{
	int source_id;
	g_return_if_fail(producer_mp3_click != NULL);
	source_id = producer_mp3_click->mp3_id;
	if (GDAM_MODEL(producer_mp3_click)->source_id && buffer_size) {
		GdamArg* arg = g_new(GdamArg, 1);
		arg->name = g_strdup("buffer_size");
		GDAM_ARG_INT(arg) = buffer_size;
		arg->type = GDAM_ARG_TYPE_INT;
		gdam_channel_append_request(
			GDAM_MODEL(producer_mp3_click)->channel,
			gdam_request_set_args(
				source_id,
				1,
				arg,
				NULL,
				NULL,
				NULL
			)
		);
	}
	producer_mp3_click->buffer_size = buffer_size;
}
