/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamequalizer.h"
#include <math.h>

#define S	1.0			/* shelving parameter */

static void compute_coefficients(GdamEqualizer* equalizer)
{
	gdouble frequency = equalizer->frequency;
	gdouble bandwidth = equalizer->bandwidth;
	gdouble db_gain = equalizer->db_gain;
	int type = equalizer->type;
	gdouble A = 0.0, omega, sn, cs, alpha, beta = 0.0;
	gdouble a0, a1, a2, b0, b1, b2;

	omega = M_2_PI * frequency / 44100.0;
	sn = sin(omega);
	cs = cos(omega);
	alpha = sn * sinh(M_LN2 / 2.0 * bandwidth * omega / sn);

	if (type == GDAM_EQUALIZER_PEAKING_EQ
	 || type == GDAM_EQUALIZER_LOW_SHELF
	 || type == GDAM_EQUALIZER_HIGH_SHELF) {
		A = exp(db_gain * M_LN10 / 40.0);
		beta  = sqrt((A * A + 1.0) / S - ((A-1.0) * (A-1.0)));
	}

	switch (type) {
		case GDAM_EQUALIZER_LOW_PASS:
			b0 =  (1.0 - cs)/2.0;
			b1 =   1.0 - cs;
			b2 =  (1.0 - cs)/2.0;
			a0 =   1.0 + alpha;
			a1 =  -2.0*cs;
			a2 =   1.0 - alpha;
			break;
		case GDAM_EQUALIZER_HIGH_PASS:
			b0 =  (1.0 + cs)/2.0;
			b1 = -(1.0 + cs);
			b2 =  (1.0 + cs)/2.0;
			a0 =   1.0 + alpha;
			a1 =  -2.0*cs;
			a2 =   1.0 - alpha;
			break;
		case GDAM_EQUALIZER_BAND_PASS:
			b0 =   alpha;
			b1 =   0;
			b2 =  -alpha;
			a0 =   1.0 + alpha;
			a1 =  -2.0*cs;
			a2 =   1.0 - alpha;
			break;
		case GDAM_EQUALIZER_NOTCH:
			b0 =   1.0;
			b1 =  -2.0*cs;
			b2 =   1.0;
			a0 =   1.0 + alpha;
			a1 =  -2.0*cs;
			a2 =   1.0 - alpha;
			break;
		case GDAM_EQUALIZER_PEAKING_EQ:
			b0 =   1.0 + alpha*A;
			b1 =  -2.0*cs;
			b2 =   1.0 - alpha*A;
			a0 =   1.0 + alpha/A;
			a1 =  -2.0*cs;
			a2 =   1.0 - alpha/A;
			break;
		case GDAM_EQUALIZER_LOW_SHELF:
			b0 =      A*( (A+1.0) - (A-1.0)*cs + beta*sn );
			b1 =  2.0*A*( (A-1.0) - (A+1.0)*cs           );
			b2 =      A*( (A+1.0) - (A-1.0)*cs - beta*sn );
			a0 =          (A+1.0) + (A-1.0)*cs + beta*sn;
			a1 =   -2.0*( (A-1.0) + (A+1.0)*cs           );
			a2 =          (A+1.0) + (A-1.0)*cs - beta*sn;
			break;
		case GDAM_EQUALIZER_HIGH_SHELF:
			b0 =      A*( (A+1.0) + (A-1.0)*cs + beta*sn );
			b1 = -2.0*A*( (A-1.0) + (A+1.0)*cs           );
			b2 =      A*( (A+1.0) + (A-1.0)*cs - beta*sn );
			a0 =          (A+1.0) - (A-1.0)*cs + beta*sn;
			a1 =    2.0*( (A-1.0) - (A+1.0)*cs           );
			a2 =          (A+1.0) - (A-1.0)*cs - beta*sn;
			break;
		default:
			gdam_debug("bad equalizer type: %d", type);
			return;
	}

	gdam_debug("a012=%g,%g,%g; b012=%g,%g,%g", a0,a1,a2,b0,b1,a2);

	equalizer->x_coeff[0] = (GdamFixed)(+b0 / a0 * GdamFixed_ONE);
	equalizer->x_coeff[1] = (GdamFixed)(+b1 / a0 * GdamFixed_ONE);
	equalizer->x_coeff[2] = (GdamFixed)(+b2 / a0 * GdamFixed_ONE);
	equalizer->y_coeff[1] = (GdamFixed)(-a1 / a0 * GdamFixed_ONE);
	equalizer->y_coeff[2] = (GdamFixed)(-a2 / a0 * GdamFixed_ONE);
	gdam_debug("equalizer: x_coeffs = [%g, %g, %g], "
			"y_coeffs = [%g, %g, %g]",
			+b0 / a0, +b1 / a0, +b2 / a0,
			0.0, -a1 / a0, -a2 / a0 );

}

static void gdam_equalizer_arg_set_frequency(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->frequency = GDAM_ARG_DOUBLE(arg);
	compute_coefficients(equalizer);
}
static void gdam_equalizer_arg_get_frequency(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = equalizer->frequency;
}
static void gdam_equalizer_arg_set_effect(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->effect = GDAM_ARG_DOUBLE(arg) * GdamFixed_ONE;
	equalizer->effect = CLAMP(equalizer->effect, 0, GdamFixed_ONE);
}
static void gdam_equalizer_arg_get_effect(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = (gdouble) equalizer->effect / GdamFixed_ONE;
}
static void gdam_equalizer_arg_set_bandwidth(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->bandwidth = GDAM_ARG_DOUBLE(arg);
	compute_coefficients(equalizer);
}
static void gdam_equalizer_arg_get_bandwidth(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = equalizer->bandwidth;
}
static void gdam_equalizer_arg_set_db_gain(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->db_gain = GDAM_ARG_DOUBLE(arg);
	compute_coefficients(equalizer);
}
static void gdam_equalizer_arg_get_db_gain(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = equalizer->db_gain;
}
static void gdam_equalizer_arg_set_mode(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	const char* mode = GDAM_ARG_STRING(arg);
	if (strcasecmp(mode, "lowpass") == 0) {
		equalizer->type = GDAM_EQUALIZER_LOW_PASS;
	} else if (strcasecmp(mode, "highpass") == 0) {
		equalizer->type = GDAM_EQUALIZER_HIGH_PASS;
	} else if (strcasecmp(mode, "bandpass") == 0) {
		equalizer->type = GDAM_EQUALIZER_BAND_PASS;
	} else if (strcasecmp(mode, "peakingeq") == 0) {
		equalizer->type = GDAM_EQUALIZER_PEAKING_EQ;
	} else if (strcasecmp(mode, "notch") == 0) {
		equalizer->type = GDAM_EQUALIZER_NOTCH;
	} else if (strcasecmp(mode, "lowshelf") == 0) {
		equalizer->type = GDAM_EQUALIZER_LOW_SHELF;
	} else if (strcasecmp(mode, "highshelf") == 0) {
		equalizer->type = GDAM_EQUALIZER_HIGH_SHELF;
	} else {
		g_warning("unknowning equalizing mode %s", mode);
		return;
	}
	compute_coefficients(equalizer);
}

static void gdam_equalizer_arg_get_mode(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	const char* mode;

	switch (equalizer->type) {
		case GDAM_EQUALIZER_LOW_PASS: mode = "lowpass"; break;
		case GDAM_EQUALIZER_HIGH_PASS: mode = "highpass"; break; 
		case GDAM_EQUALIZER_BAND_PASS: mode = "bandpass"; break;
		case GDAM_EQUALIZER_PEAKING_EQ: mode = "peakingeq"; break;
		case GDAM_EQUALIZER_NOTCH: mode = "notch"; break;
		case GDAM_EQUALIZER_LOW_SHELF: mode = "lowshelf"; break;
		case GDAM_EQUALIZER_HIGH_SHELF: mode = "highshelf"; break;
		default: mode = "unknown";
	}
	GDAM_ARG_STRING(arg) = g_strdup(mode);
}

static inline void process_sample(GdamEqualizer* equalizer, GdamSample* in_out)
{
	int y = 0;
	int in0, in1;
	GdamSample* x_arr = equalizer->x;
	GdamSample* y_arr = equalizer->y;
	GdamFixed*  x_coeff = equalizer->x_coeff;
	GdamFixed*  y_coeff = equalizer->y_coeff;
	GdamSample  out_eq;

	in0 = (in_out->s[0] * equalizer->effect / 2) / (GdamFixed_ONE/2);
	in1 = (in_out->s[1] * equalizer->effect / 2) / (GdamFixed_ONE/2);

	y = ((x_coeff[0] >> 4) * in0)
          + ((x_coeff[1] >> 4) * (int)(x_arr[0].s[0]))
          + ((x_coeff[2] >> 4) * (int)(x_arr[1].s[0]))
          + ((y_coeff[1] >> 4) * (int)(y_arr[0].s[0]))
          + ((y_coeff[2] >> 4) * (int)(y_arr[1].s[0]));
	y >>= (GdamFixed_LOG_ONE-4);
	out_eq.s[0] = CLAMP(y, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);

	y += ((int)in_out->s[0] - in0);
	if (y < GDAM_SAMPLE_MIN) y = GDAM_SAMPLE_MIN;
	else if (y > GDAM_SAMPLE_MAX) y = GDAM_SAMPLE_MAX;
	in_out->s[0] = (gint16) y;

	y = ((x_coeff[0] >> 4) * in1)
          + ((x_coeff[1] >> 4) * (int)(x_arr[0].s[1]))
          + ((x_coeff[2] >> 4) * (int)(x_arr[1].s[1]))
          + ((y_coeff[1] >> 4) * (int)(y_arr[0].s[1]))
          + ((y_coeff[2] >> 4) * (int)(y_arr[1].s[1]));
	y >>= (GdamFixed_LOG_ONE-4);
	out_eq.s[1] = CLAMP(y, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);

	y += ((int)in_out->s[1] - in1);
	if (y < GDAM_SAMPLE_MIN) y = GDAM_SAMPLE_MIN;
	else if (y > GDAM_SAMPLE_MAX) y = GDAM_SAMPLE_MAX;
	in_out->s[1] = (gint16) y;

	x_arr[1] = x_arr[0];
	y_arr[1] = y_arr[0];
	x_arr[0].s[0] = CLAMP(in0, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);
	x_arr[0].s[1] = CLAMP(in1, GDAM_SAMPLE_MIN, GDAM_SAMPLE_MAX);
	y_arr[0] = out_eq;
}

static int gdam_equalizer_read(GdamSource* source, GdamSample* buf, int n)
{
	GdamSource* filtered = GDAM_FILTER(source)->filtered;
	GdamEqualizer* equalizer = GDAM_EQUALIZER(source);
	int i, num_read;
	if (filtered == NULL)
		return 0;
	num_read = gdam_source_read(filtered, buf, n);
	for (i = 0; i < num_read; i++)
		process_sample(equalizer, buf++);
	return num_read;
}

static void gdam_equalizer_class_init(GtkObjectClass* object_class)
{
	GDAM_SOURCE_CLASS(object_class)->read = gdam_equalizer_read;
	gdam_arg_register(object_class,
	                  "frequency",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_frequency,
			  gdam_equalizer_arg_get_frequency,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "effect",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_effect,
			  gdam_equalizer_arg_get_effect,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "bandwidth",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_bandwidth,
			  gdam_equalizer_arg_get_bandwidth,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "db_gain",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_db_gain,
			  gdam_equalizer_arg_get_db_gain,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "mode",
			  GDAM_ARG_TYPE_STRING,
			  gdam_equalizer_arg_set_mode,
			  gdam_equalizer_arg_get_mode,
			  GDAM_ARG_VITAL);
}
static void gdam_equalizer_init(GtkObject* object)
{
	GdamEqualizer* equalizer = (GdamEqualizer*)object;
	equalizer->type = GDAM_EQUALIZER_HIGH_PASS;
	equalizer->frequency = 440.0;
	equalizer->bandwidth = 1.0;
	equalizer->db_gain = 1.0;

	equalizer->effect = GdamFixed_ONE;

	/* Never used, but we zero it out initially then never touch it... */
	equalizer->y_coeff[0] = 0;
	equalizer->x_coeff[0] = GdamFixed_ONE;

	//compute_coefficients(equalizer);
}

GtkType gdam_equalizer_get_type()
{
	static GtkType equalizer_type = 0;
	if (!equalizer_type) {
		static const GtkTypeInfo equalizer_info = {
			"GdamEqualizer",
			sizeof(GdamEqualizer),
			sizeof(GdamFilterClass),
			(GtkClassInitFunc) gdam_equalizer_class_init,
			(GtkObjectInitFunc) gdam_equalizer_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		equalizer_type = gtk_type_unique(GDAM_TYPE_FILTER,
		                                 &equalizer_info);
	}
	return equalizer_type;
}
